/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro
 Maria del Carmen Sánchez Medrano

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
 package org.javafmi.kernel.v2rc2.nativeaccess;

import com.sun.jna.Memory;
import com.sun.jna.Native;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.PointerByReference;
import org.javafmi.exceptions.MethodExecutionException;
import org.javafmi.kernel.operatingsystem.NativeBoolean;
import org.javafmi.kernel.operatingsystem.SharedMemoryAllocator;
import org.javafmi.kernel.v2.nativeaccess.CallbackFunctionsStructure;

import java.nio.ByteBuffer;
import java.nio.DoubleBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;

public class FmiV2RC2 {

    private final FmiV2RC2Interface fmiV2RC2Interface;
    private final CallbackFunctionsStructure.ByReference callbacks;
    private final List<Pointer> objectsInMemory;
    private final SharedMemoryAllocator sharedMemoryAllocator;
    private Pointer component;


    public FmiV2RC2(String libraryPath) {
        this.objectsInMemory = new ArrayList<>();
        this.callbacks = new CallbackFunctionsStructure.ByReference(objectsInMemory);
        this.fmiV2RC2Interface = (FmiV2RC2Interface) Native.loadLibrary(libraryPath, FmiV2RC2Interface.class);
        this.sharedMemoryAllocator = new SharedMemoryAllocator();
    }

    public void instantiate(String modelName, int fmiType, String guid) {
        component = fmiV2RC2Interface.fmi2Instantiate(
                modelName,
                fmiType,
                guid,
                "file:///",
                callbacks,
                false,
                false
        );
        if(component == null) throw new MethodExecutionException("fmi2Instantiate returns null");
    }

    public void freeInstance() {
        fmiV2RC2Interface.fmi2FreeInstance(component);
        for(int i = objectsInMemory.size() - 1; i >= 0; i--) objectsInMemory.remove(i);
    }

    public int setDebugLogging(boolean loggingOn, String[] categories) {
        return fmiV2RC2Interface.fmi2SetDebugLogging(component, loggingOn, categories.length, categories);
    }

    public int setupExperiment(boolean toleranceDefined, double tolerance, double startTime, boolean stopTimeDefined, double stopTime) {
        return fmiV2RC2Interface.fmi2SetupExperiment(component, toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime);
    }

    public int enterInitializationMode() {
        return fmiV2RC2Interface.fmi2EnterInitializationMode(component);
    }

    public int exitInitializationMode() {
        return fmiV2RC2Interface.fmi2ExitInitializationMode(component);
    }

    public int terminate() {
        return fmiV2RC2Interface.fmi2Terminate(component);
    }

    public int reset() {
        return fmiV2RC2Interface.fmi2Reset(component);
    }

    public double getReal(int[] valueReferences) {
        DoubleBuffer sharedDouble = sharedMemoryAllocator.getSharedDouble();
        fmiV2RC2Interface.fmi2GetReal(component, valueReferences, valueReferences.length, sharedDouble);
        return sharedDouble.get(0);
    }

    public int getInteger(int[] valueReferences) {
        IntBuffer sharedInteger = sharedMemoryAllocator.getSharedInteger();
        fmiV2RC2Interface.fmi2GetInteger(component, valueReferences, valueReferences.length, sharedInteger);
        return sharedInteger.get(0);
    }

    public boolean getBoolean(int[] valueReferences) {
        IntBuffer sharedInteger = sharedMemoryAllocator.getSharedInteger();
        fmiV2RC2Interface.fmi2GetBoolean(component, valueReferences, valueReferences.length, sharedInteger);
        return NativeBoolean.getBooleanFor((byte) sharedInteger.get(0));
    }

    public String getString(int[] valueReferences) {
        PointerByReference pointerByReference = new PointerByReference();
        fmiV2RC2Interface.fmi2GetString(component, valueReferences, valueReferences.length, pointerByReference);
        return pointerByReference.getValue().getString(0);
    }

    public int setReal(int[] valueReferences, double[] doubleValues) {
        DoubleBuffer sharedDouble = DoubleBuffer.allocate(doubleValues.length);
        for(int i = 0; i < doubleValues.length; i++) sharedDouble.put(i, doubleValues[i]);
        return fmiV2RC2Interface.fmi2SetReal(component, valueReferences, valueReferences.length, sharedDouble);
    }

    public int setInteger(int[] valueReferences, int[] intValues) {
        IntBuffer sharedInt = IntBuffer.allocate(intValues.length);
        for(int i = 0; i < intValues.length; i++) sharedInt.put(i, intValues[i]);
        return fmiV2RC2Interface.fmi2SetInteger(component, valueReferences, valueReferences.length, sharedInt);
    }

    public int setBoolean(int[] valueReferences, boolean[] booleanValues) {
        IntBuffer sharedInt = IntBuffer.allocate(booleanValues.length);
        for(int i = 0; i < booleanValues.length; i++) sharedInt.put(i, NativeBoolean.getValueFor(booleanValues[i]));
        return fmiV2RC2Interface.fmi2SetBoolean(component, valueReferences, valueReferences.length, sharedInt);
    }

    public int setString(int[] valueReferences, String[] stringValues) {
        PointerByReference pointerByReference = new PointerByReference();
        Pointer reference = new Memory(stringValues.length).share(0);
        for(int i = 0; i < stringValues.length; i++) reference.setString(i, stringValues[i]);
        pointerByReference.setValue(reference);
        return fmiV2RC2Interface.fmi2SetString(component, valueReferences, valueReferences.length, pointerByReference);
    }

    public Pointer getFMUState(Pointer state) {
        PointerByReference pointerByReference = new PointerByReference(state);
        fmiV2RC2Interface.fmi2GetFMUstate(component, pointerByReference);
        return pointerByReference.getValue();
    }

    public int setFMUState(Pointer pointer) {
        return fmiV2RC2Interface.fmi2SetFMUstate(component, pointer);
    }

    public int freeFMUState(Pointer state) {
        return fmiV2RC2Interface.fmi2FreeFMUstate(component, state);
    }

    public int serializedFMUStateSize(Pointer state) {
        IntBuffer intBuffer = IntBuffer.allocate(1);
        fmiV2RC2Interface.fmi2SerializedFMUstateSize(component, state, intBuffer);
        return intBuffer.get(0);
    }

    public byte[] serializeFMUState(Pointer state) {
        int size = serializedFMUStateSize(state);
        ByteBuffer byteBuffer = ByteBuffer.allocate(size);
        fmiV2RC2Interface.fmi2SerializeFMUstate(component, state, byteBuffer, size);
        return byteBuffer.array();
    }

    public Pointer deSerializeFMUState(byte[] serializedState) {
        PointerByReference pointerByReference = new PointerByReference();
        fmiV2RC2Interface.fmi2DeSerializeFMUstate(component, serializedState, serializedState.length, pointerByReference);
        return pointerByReference.getValue();
    }

    public double[] getDirectionalDerivative(int[] unknownVariablesValueReferences, int[] knownVariablesValueReferences, double[] knownDifferential) {
        DoubleBuffer sharedDoubles = sharedMemoryAllocator.shareDoubles(knownDifferential.length);
        fmiV2RC2Interface.fmi2GetDirectionalDerivative(component, unknownVariablesValueReferences, unknownVariablesValueReferences.length, knownVariablesValueReferences, knownVariablesValueReferences.length, knownDifferential, sharedDoubles);
        return sharedDoubles.array();
    }

    public int doStep(double currentSimulationTime, double stepSize, boolean noSetFmuStatePriorToCurrentPoint) {
        return fmiV2RC2Interface.fmi2DoStep(component, currentSimulationTime, stepSize, noSetFmuStatePriorToCurrentPoint);
    }

    public int cancelStep() {
        return fmiV2RC2Interface.fmi2CancelStep(component);
    }
}
